/* This file is part of Cloudy and is copyright (C)1978-2014 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
/*CoolIron compute iron cooling */
/*Fe2_cooling compute cooling due to FeII emission */
#include "cddefines.h"
#include "dense.h"
#include "taulines.h"
#include "phycon.h"
#include "iso.h"
#include "conv.h"
#include "cooling.h"
#include "trace.h"
#include "hydrogenic.h"
#include "thermal.h"
#include "atoms.h"
#include "atom_feii.h"
#include "lines_service.h"
#include "rt.h"

/*Fe2_cooling compute cooling due to FeII emission */
STATIC void Fe2_cooling( void )
{
	const int NLFE2 = 6;
	long int i, j;
	int nNegPop;

	static double **AulPump,
		**CollRate,
		**AulEscp,
		**col_str ,
		**AulDest, 
		*depart,
		*pops,
		*destroy,
		*create;

	static bool lgFirst=true;
	bool lgZeroPop;

	/* stat weights for Fred's 6 level model FeII atom */
	static double gFe2[NLFE2]={1.,1.,1.,1.,1.,1.};
	/* excitation energies (Kelvin) for Fred's 6 level model FeII atom */
	static double ex[NLFE2]={0.,3.32e4,5.68e4,6.95e4,1.15e5,1.31e5};

	/* these are used to only evaluated FeII one time per temperature, zone, 
	 * and abundance */
	static double TUsed = 0.; 
	static double AbunUsed = 0.;
	/* remember which zone last evaluated with */
	static long int nZUsed=-1,
		/* make sure at least two calls per zone */
		nCall=0;

	DEBUG_ENTRY( "Fe2_cooling()" );

	/* return if nothing to do */
	if( dense.xIonDense[ipIRON][1] == 0. )
	{
		/* zero abundance, do nothing */
		/* heating cooling and derivative from large atom */
		FeII.Fe2_large_cool = 0.;
		FeII.Fe2_large_heat = 0.;
		FeII.ddT_Fe2_large_cool = 0.;

		/* cooling and derivative from simple UV atom */
		FeII.Fe2_UVsimp_cool = 0.;
		FeII.ddT_Fe2_UVsimp_cool = 0.;

		/* now zero out intensities of all FeII lines and level populations */
		FeIIIntenZero();
		return;
	}

	/* evaluate FeII model atom, by default only the lowest 16 levels
	 * but number of levels can be increased with atom feii command */

	/* totally reevaluate FeII atom if new zone, or cooling is significant
	 * and temperature changed, we are in search phase,
	 * lgSlow option set true with atom FeII slow, forces constant
	 * evaluation of atom */
	if( FeII.lgSlow || !lgConvBaseHeatTest ||
		conv.lgFirstSweepThisZone || conv.lgLastSweepThisZone ||
	    /* check whether things have changed on later calls */
	    ( !fp_equal( phycon.te, TUsed ) && fabs(FeII.Fe2_large_cool/thermal.ctot) > 0.002 &&  
	      fabs(dense.xIonDense[ipIRON][1]-AbunUsed)/SDIV(AbunUsed) > 0.002 ) ||
	    ( !fp_equal( phycon.te, TUsed ) && fabs(FeII.Fe2_large_cool/thermal.ctot) > 0.01) )
	{

		if( conv.lgFirstSweepThisZone )
			/* first call this zone set nCall to zero*/
			nCall = 0;
		else
			/* not first call, increment, check above to make sure at least
			 * two evaluations */
			++nCall;

		/* option to trace convergence and FeII calls */
		if( trace.nTrConvg >= 5 )
		{
			fprintf( ioQQQ, "        CoolIron5 calling FeIILevelPops since ");
			if( conv.lgFirstSweepThisZone )
			{
				fprintf( ioQQQ, 
					"first sweep this zone." );
			}
			else if( ! fp_equal( phycon.te, TUsed ) )
			{
				fprintf( ioQQQ, 
					"temperature changed, old new are %g %g, nCall %li ", 
					TUsed, phycon.te , nCall);
			}
			else if( nzone != nZUsed )
			{
				fprintf( ioQQQ, 
					"new zone, nCall %li ", nCall );
			}
			else if( FeII.lgSlow )
			{
				fprintf( ioQQQ, 
					"FeII.lgSlow set  %li", nCall );
			}
			else if( conv.lgSearch )
			{
				fprintf( ioQQQ, 
					" in search phase  %li", nCall );
			}
			else if( nCall < 2 )
			{
				fprintf( ioQQQ, 
					"not second nCall %li " , nCall );
			}
			else if( ! fp_equal( phycon.te, TUsed ) && FeII.Fe2_large_cool/thermal.ctot > 0.001 )
			{
				fprintf( ioQQQ, 
					"temp or cooling changed, new are %g %g nCall %li ", 
					phycon.te, FeII.Fe2_large_cool, nCall );
			}
			else
			{
				fprintf(ioQQQ, "????");
			}
			fprintf(ioQQQ, "\n");
		}

		/* remember parameters for current conditions */
		TUsed = phycon.te;
		AbunUsed = dense.xIonDense[ipIRON][1];
		nZUsed = nzone;

		/* this print turned on with atom FeII print command */
		if( FeII.lgPrint )
		{
			fprintf(ioQQQ,
				" FeIILevelPops called zone %4li te %5f abun %10e c(fe/tot):%6f nCall %li\n", 
				nzone,phycon.te,AbunUsed,FeII.Fe2_large_cool/thermal.ctot,nCall);
		}

		/* this solves the multi-level problem, 
		 * sets FeII.Fe2_large_cool, 
				FeII.Fe2_large_heat, & 
				FeII.ddT_Fe2_large_cool 
				but does nothing with them */

		// line RT update, followed by level population solver
		FeII_RT_Make(RT_line_one_escape);
		FeIILevelPops();
		{
			enum{DEBUG_LOC=false};
			if( DEBUG_LOC && iteration > 1 && nzone >=4 )
			{
				fprintf(ioQQQ,"DEBUG1\t%li\t%.3e\t%.3e\t%.3e\t%.3e\t%.3e\t%.3e\t%.3e\n", 
					nzone,
					phycon.te,
					dense.gas_phase[ipHYDROGEN],
					dense.eden,
					FeII.Fe2_large_cool , 
					FeII.ddT_Fe2_large_cool ,
					FeII.Fe2_large_cool/dense.eden/dense.gas_phase[ipHYDROGEN] , 
					thermal.ctot );
			}
		}

		if( trace.nTrConvg >= 5 || FeII.lgPrint)
		{
			/* spacing needed to get proper trace convergence output */
			fprintf( ioQQQ, "           FeIILevelPops5 returned cool=%.2e heat=%.2e derivative=%.2e\n",
					FeII.Fe2_large_cool,FeII.Fe2_large_heat ,FeII.ddT_Fe2_large_cool);
		}

	}
	else if( ! fp_equal( dense.xIonDense[ipIRON][1], AbunUsed ) )
	{
		realnum ratio;
		/* this branch, same zone and temperature, but small change in abundance, so just
		 * rescale cooling and derivative by this change.  assumption is that very small changes
		 * in abundance occurs as ots rates damp out */
		if( trace.nTrConvg >= 5 )
		{
			fprintf( ioQQQ, 
				"       CoolIron rescaling FeIILevelPops since small change, CFe2=%.2e CTOT=%.2e\n",
				FeII.Fe2_large_cool,thermal.ctot);
		}
		ratio = dense.xIonDense[ipIRON][1]/AbunUsed;
		FeII.Fe2_large_cool *= ratio;
		FeII.ddT_Fe2_large_cool *= ratio;
		FeII.Fe2_large_heat *= ratio;
		AbunUsed = dense.xIonDense[ipIRON][1];
	}
	else
	{
		/* this is case where temp is unchanged, so heating and cooling same too */
		if( trace.nTrConvg >= 5 )
		{
			fprintf( ioQQQ, "       CoolIron NOT calling FeIILevelPops\n");
		}
	}

	/* now update total cooling and its derivative 
	 * all paths flow through here */
	/* FeII.Fecool = FeII.Fe2_large_cool; */
	CoolAdd("Fe 2",0,MAX2(0.,FeII.Fe2_large_cool));

	/* add negative cooling to heating stack */
	thermal.setHeating(25,27, MAX2(0.,FeII.Fe2_large_heat));

	/* counts as heating derivative if negative cooling */
	if( FeII.Fe2_large_cool > 0. )
	{
		/* >>chng 01 mar 16, add factor of 3 due to conv problems after changing damper */
		thermal.dCooldT += 3.*FeII.ddT_Fe2_large_cool;
	}

	if( trace.lgTrace && trace.lgCoolTr )
	{
		fprintf( ioQQQ, " Large FeII returns te, cooling, dc=%11.3e%11.3e%11.3e\n", 
		  phycon.te, FeII.Fe2_large_cool, FeII.ddT_Fe2_large_cool );
	}

	/* >>chng 05 nov 29, still do simple UV atom if only ground term is done */
	if( !FeII.lgFeIILargeOn )
	{

		/* following treatment of Fe II follows
		 * >>refer	fe2	model	Wills, B.J., Wills, D., Netzer, H. 1985, ApJ, 288, 143
		 * all elements are used, and must be set to zero if zero */

		/* set up space for simple  model of UV FeII emission */
		if( lgFirst )
		{
			/* will never do this again in this core load */
			lgFirst = false;
			/* allocate the 1D arrays*/
			pops = (double *)MALLOC( sizeof(double)*(NLFE2) );
			create = (double *)MALLOC( sizeof(double)*(NLFE2) );
			destroy = (double *)MALLOC( sizeof(double)*(NLFE2) );
			depart = (double *)MALLOC( sizeof(double)*(NLFE2) );
			/* create space for the 2D arrays */
			AulPump = ((double **)MALLOC((NLFE2)*sizeof(double *)));
			CollRate = ((double **)MALLOC((NLFE2)*sizeof(double *)));
			AulDest = ((double **)MALLOC((NLFE2)*sizeof(double *)));
			AulEscp = ((double **)MALLOC((NLFE2)*sizeof(double *)));
			col_str = ((double **)MALLOC((NLFE2)*sizeof(double *)));

			for( i=0; i < NLFE2; ++i )
			{
				AulPump[i] = ((double *)MALLOC((NLFE2)*sizeof(double )));
				CollRate[i] = ((double *)MALLOC((NLFE2)*sizeof(double )));
				AulDest[i] = ((double *)MALLOC((NLFE2)*sizeof(double )));
				AulEscp[i] = ((double *)MALLOC((NLFE2)*sizeof(double )));
				col_str[i] = ((double *)MALLOC((NLFE2)*sizeof(double )));
			}
		}

		/*zero out all arrays, then check that upper diagonal remains zero below */
		for( i=0; i < NLFE2; i++ )
		{
			create[i] = 0.;
			destroy[i] = 0.;
			for( j=0; j < NLFE2; j++ )
			{
				/*data[j][i] = 0.;*/
				col_str[j][i] = 0.;
				AulEscp[j][i] = 0.;
				AulDest[j][i] = 0.;
				AulPump[j][i] = 0.;
			}
		}

		/* now put in real data for lines */
		AulEscp[1][0] = 1.;
		AulEscp[2][0] = ( TauLines[ipTuv3].Emis().Pesc_total())*TauLines[ipTuv3].Emis().Aul();
		AulDest[2][0] = TauLines[ipTuv3].Emis().Pdest()*TauLines[ipTuv3].Emis().Aul();
		AulPump[0][2] = TauLines[ipTuv3].Emis().pump();

		AulEscp[5][0] = (TauLines[ipTFe16].Emis().Pesc_total())*TauLines[ipTFe16].Emis().Aul();
		AulDest[5][0] = TauLines[ipTFe16].Emis().Pdest()*TauLines[ipTFe16].Emis().Aul();
		/* continuum pumping of n=6 */
		AulPump[0][5] = TauLines[ipTFe16].Emis().pump();
		/* Ly-alpha pumping */

		double PumpLyaFeII = iso_sp[ipH_LIKE][ipHYDROGEN].st[ipH2p].Pop()*
			iso_sp[ipH_LIKE][ipHYDROGEN].trans(ipH2p,ipH1s).Emis().Aul()*
			hydro.dstfe2lya/SDIV(dense.xIonDense[ipIRON][1]);

		AulPump[0][5] += PumpLyaFeII;

		AulEscp[2][1] = (TauLines[ipTr48].Emis().Pesc_total())*TauLines[ipTr48].Emis().Aul();
		AulDest[2][1] = TauLines[ipTr48].Emis().Pdest()*TauLines[ipTr48].Emis().Aul();
		AulPump[1][2] = TauLines[ipTr48].Emis().pump();

		AulEscp[5][1] = (TauLines[ipTFe26].Emis().Pesc_total())*TauLines[ipTFe26].Emis().Aul();
		AulDest[5][1] = TauLines[ipTFe26].Emis().Pdest()*TauLines[ipTFe26].Emis().Aul();
		AulPump[1][5] = TauLines[ipTFe26].Emis().pump();

		AulEscp[3][2] = (TauLines[ipTFe34].Emis().Pesc_total())*TauLines[ipTFe34].Emis().Aul();
		AulDest[3][2] = TauLines[ipTFe34].Emis().Pdest()*TauLines[ipTFe34].Emis().Aul();
		AulPump[2][3] = TauLines[ipTFe34].Emis().pump();

		AulEscp[4][2] = (TauLines[ipTFe35].Emis().Pesc_total())*TauLines[ipTFe35].Emis().Aul();
		AulDest[4][2] = TauLines[ipTFe35].Emis().Pdest()*TauLines[ipTFe35].Emis().Aul();
		AulPump[2][4] = TauLines[ipTFe35].Emis().pump();

		AulEscp[5][3] = (TauLines[ipTFe46].Emis().Pesc_total())*TauLines[ipTFe46].Emis().Aul();
		AulDest[5][3] = TauLines[ipTFe46].Emis().Pdest()*TauLines[ipTFe46].Emis().Aul();
		AulPump[3][5] = TauLines[ipTFe46].Emis().pump();

		AulEscp[5][4] = (TauLines[ipTFe56].Emis().Pesc_total())*TauLines[ipTFe56].Emis().Aul();
		AulDest[5][4] = TauLines[ipTFe56].Emis().Pdest()*TauLines[ipTFe56].Emis().Aul();
		AulPump[4][5] = TauLines[ipTFe56].Emis().pump();

		// Set AulPump Dx Rates
		for( int low=0; low < NLFE2-1; low++ )
		{
			for( int ihi=low+1; ihi < NLFE2; ihi++ )
			{
				AulPump[ihi][low] = AulPump[low][ihi]*gFe2[low]/gFe2[ihi];
			}
		}

		/* these are collision strengths */
		col_str[1][0] = 1.;
		col_str[2][0] = 12.;
		col_str[3][0] = 1.;
		col_str[4][0] = 1.;
		col_str[5][0] = 12.;
		col_str[2][1] = 6.;
		col_str[3][1] = 1.;
		col_str[4][1] = 1.;
		col_str[5][1] = 12.;
		col_str[3][2] = 6.;
		col_str[4][2] = 12.;
		col_str[5][2] = 1.;
		col_str[4][3] = 1.;
		col_str[5][3] = 12.;
		col_str[5][4] = 6.;

		/*void atom_levelN(long,long,realnum,double[],double[],double[],double*,
		double*,double*,long*,realnum*,realnum*,STRING,int*);*/
		atom_levelN(NLFE2,
			dense.xIonDense[ipIRON][1],
			gFe2,
			ex,
			'K',
			pops,
			depart,
			AulEscp ,
			col_str,
			AulDest,
			AulPump,
			CollRate,
			create,
			destroy,
			false,/* say atom_levelN should evaluate coll rates from cs */
			/*&&ipdest,*/
			&FeII.Fe2_UVsimp_cool,
			&FeII.ddT_Fe2_UVsimp_cool,
			"FeII",
			&nNegPop,
			&lgZeroPop,
			false );

		/* nNegPop positive if negative pops occurred, negative if too cold */
		if( nNegPop > 0 /*negative if too cold - that is not negative and is OK */ )
		{
			fprintf(ioQQQ," PROBLEM, atom_levelN returned negative population for simple UV FeII.\n");
		}

		/* add heating - cooling by this process */;
		CoolAdd("Fe 2",0,MAX2(0.,FeII.Fe2_UVsimp_cool));
		thermal.setHeating(25,27, MAX2(0.,-FeII.Fe2_UVsimp_cool));
		thermal.dCooldT += FeII.ddT_Fe2_UVsimp_cool;

		/* LIMLEVELN is the dim of the PopLevels vector */
		ASSERT( NLFE2 <= LIMLEVELN );
		for( i=0; i < NLFE2; ++i )
		{
			atoms.PopLevels[i] = pops[i];
			atoms.DepLTELevels[i] = depart[i];
		}

		(*TauLines[ipTuv3].Lo()).Pop() = pops[0];
		(*TauLines[ipTuv3].Hi()).Pop() = pops[2];
		TauLines[ipTuv3].Emis().PopOpc() = (pops[0] - pops[2]);
		set_xIntensity( TauLines[ipTuv3] );

		(*TauLines[ipTr48].Lo()).Pop() = pops[1];
		(*TauLines[ipTr48].Hi()).Pop() = pops[2];
		TauLines[ipTr48].Emis().PopOpc() = (pops[1] - pops[2]);
		set_xIntensity( TauLines[ipTr48] );

		FeII.for7 = pops[1]*AulEscp[1][0]*4.65e-12;

		(*TauLines[ipTFe16].Lo()).Pop() = pops[0];
		(*TauLines[ipTFe16].Hi()).Pop() = pops[5];
		/* Lyman alpha optical depths are not known on first iteration,
		 * inward optical depths used, so line trapping overestimated,
		 * this can cause artificial maser in FeII - prevent by not
		 * including stimulated emission correction on first iteration */
		TauLines[ipTFe16].Emis().PopOpc() = (pops[0] - pops[5]);
		set_xIntensity( TauLines[ipTFe16] );

		(*TauLines[ipTFe26].Lo()).Pop() = pops[1];
		(*TauLines[ipTFe26].Hi()).Pop() = pops[5];
		TauLines[ipTFe26].Emis().PopOpc() = (pops[1] - pops[5]);
		set_xIntensity( TauLines[ipTFe26] );

		(*TauLines[ipTFe34].Lo()).Pop() = pops[2];
		(*TauLines[ipTFe34].Hi()).Pop() = pops[3];
		TauLines[ipTFe34].Emis().PopOpc() = (pops[2] - pops[3]);
		set_xIntensity( TauLines[ipTFe34] );

		(*TauLines[ipTFe35].Lo()).Pop() = pops[2];
		(*TauLines[ipTFe35].Hi()).Pop() = pops[4];
		TauLines[ipTFe35].Emis().PopOpc() = (pops[2] - pops[4]);
		set_xIntensity( TauLines[ipTFe35] );

		(*TauLines[ipTFe46].Lo()).Pop() = pops[3];
		(*TauLines[ipTFe46].Hi()).Pop() = pops[5];
		TauLines[ipTFe46].Emis().PopOpc() = (pops[3] - pops[5]);
		set_xIntensity( TauLines[ipTFe46] );

		(*TauLines[ipTFe56].Lo()).Pop() = pops[4];
		(*TauLines[ipTFe56].Hi()).Pop() = pops[5];
		TauLines[ipTFe56].Emis().PopOpc() = (pops[4] - pops[5]);
		set_xIntensity( TauLines[ipTFe56] );

		/* Jack's funny FeII lines, data from 
		 * >>refer	fe2	energy	Johansson, S., Brage, T., Leckrone, D.S., Nave, G. &
		 * >>refercon Wahlgren, G.M. 1995, ApJ 446, 361 */
		PutCS(10.,TauLines[ipT191]);
		atom_level2(TauLines[ipT191]);
	}

	{
		/*@-redef@*/
		enum{DEBUG_LOC=false};
		/*@+redef@*/
		if( DEBUG_LOC && iteration > 1 && nzone >=4 )
		{
			fprintf(ioQQQ,"DEBUG2\t%.2e\t%.2e\t%.2e\n",
				phycon.te,
				FeII.Fe2_large_cool , 
				FeII.Fe2_UVsimp_cool );
		}
	}

	return;

}

/*CoolIron - calculation total cooling due to Fe */
void CoolIron(void)
{
	DEBUG_ENTRY( "CoolIron()" );

	/* >>chng 05 dec 03, move Fe2 FeII Fe II cooling into separate routine */
	Fe2_cooling();

	return;
}
