/* This file is part of Cloudy and is copyright (C)1978-2014 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
#include "cddefines.h"
#include "lines.h"
#include "cddrive.h"
#include "prt.h"
#include "lines_service.h"

t_LineSave LineSave;
/* these are the definitions of the line save arrays in lines.h */

void t_LineSave::zero()
{
	DEBUG_ENTRY(" t_LineSave::zero()");

	/* index within the line in the line stack 
	 * default is Hbeta total - the third line in the stack
	 * 0th is a zero for sanity, 1st is unit, 2nd is a comment */
	/* >>chng 02 apr 22 from 2 to 3 since added unit at 1 */
	/* >>chng 06 mar 11, from 3 to -1 will now set to "H  1" 4861 */
	ipNormWavL = -1;
	WavLNorm = 4861.36f;
	lgNormSet = false;
	sig_figs = 6;

	/* the label for the normalization line */
	strcpy( chNormLab, "    " );

	/* The scale factor for the normalization line */
	ScaleNormLine = 1.;

}

void LinSv::prt(FILE* ioPUN) const
{
	DEBUG_ENTRY(" LinSv::prt()");
	fprintf(ioPUN,"%s",label().c_str());
}

string LinSv::label() const
{
	DEBUG_ENTRY(" LinSv::label()");
	string val = chALab();
	val += " ";
	char buf[100];
	sprt_wl(buf, wavelength());
	val += buf;
	return val;
}

string LinSv::biglabel() const
{
	DEBUG_ENTRY(" LinSv::biglabel()");
	string val = label();
	if (m_tr.associated())
	{
		val += " ";
		val += "[";
		val += (*m_tr.Hi()).chConfig();
		val += "->";
		val += (*m_tr.Lo()).chConfig();
		val += "]";
	}
	return val;
}

bool LinSv::isCat(const char *s) const
{
	DEBUG_ENTRY(" LinSv::isCat()");
	return strncmp(chALab(),s,strlen(s)) == 0;
}

void LinSv::chALabSet(const char *that)
{
	/* check that null is correct, string overruns have 
	 * been a problem in the past */
	ASSERT( (int) strlen( that )< NCHLAB );
	strncpy(m_chALab,that,NCHLAB-1);
	m_chALab[NCHLAB-1] = '\0';
	strncpy(m_chCLab,that,NCHLAB-1);
	m_chCLab[NCHLAB-1] = '\0';
	caps(m_chCLab);
}

void LinSv::init(char chSumTyp, const char *chComment, const char *label,
					  bool lgAdd, realnum wavelength, const TransitionProxy& tr)
{
	/* first call to stuff lines in array, confirm that label is one of
	 * the four correct ones */
	ASSERT( (chSumTyp == 'c') || (chSumTyp == 'h') || (chSumTyp == 'i') || (chSumTyp == 'r' ) );
	ASSERT(!m_tr.associated());
	/* then save it into array */
	m_chSumTyp = chSumTyp;
	emslinZero();
	m_chComment = chComment;
	chALabSet( label );
	
	if (!lgAdd)
	{
		// number of lines OK, set parameters for first pass
		// negative wavelength means it is just label, possibly not correct
		wavelength = fabs(wavelength);
	}
	SumLineZero();
	m_wavelength = wavelength;
	m_tr = tr;
}

void LinSv::addComponent(const char* species,const double wavelength)
{
	if ( LineSave.ipass == 0 )
	{
		long id = LineSave.findline(species,wlAirVac(wavelength));
		m_component.push_back(id);
		if (m_component.size() == 1)
			m_chComment += ": ";
		else
			m_chComment += "+";
		if (id > 0)
		{
			m_chComment += "\""+LineSave.lines[id].label()+"\"";
		}
		else
		{
			m_chComment += "N/A";
			fprintf( ioQQQ, "Error: A component to line blend \"%s\" %.3f was not identified.\n",
					species, wavelength );
		}
	}
}
string LinSv::chComment() const
{
	return m_chComment;
}

long t_LineSave::findline(const char *chLabel, realnum wavelength)
{
	DEBUG_ENTRY(" t_LineSave::findline()");
	
	ASSERT( nsum >= 0);
	if (nsum == 0)
		return -1;

	bool lgDEBUG = false;
	
	/* check that chLabel[4] is null - supposed to be 4 char + end */
	if( chLabel[NCHLAB-1] != '\0' || strlen(chLabel) != NCHLAB-1 )
	{
		fprintf( ioQQQ, " findline called with insane chLabel (between quotes) \"%s\", must be exactly %d characters long.\n",
					chLabel, NCHLAB-1 );
		return -2;
	}
 
	char chCARD[INPUT_LINE_LENGTH];
	strcpy( chCARD, chLabel );

	/* make sure chLabel is all caps */
	caps(chCARD);/* convert to caps */

	/* this replaces tabs with spaces. */
	/* \todo	2 keep this in, do it elsewhere, just warn and bail? */
	for( long j=0; j<NCHLAB-1; j++ )
	{
		if( chCARD[j] == '\t' )
		{
			chCARD[j] = ' ';
		}
	}

	/* get the error associated with specified significant figures; add
	 * FLT_EPSILON*wavelength to broaden bounds enough to allow for
	 * cancellation error
	 */
	realnum errorwave = WavlenErrorGet( wavelength ) + FLT_EPSILON*wavelength, 
		smallest_error=BIGFLOAT,
		smallest_error_w_correct_label=BIGFLOAT;

	long int j, index_of_closest=LONG_MIN,
		index_of_closest_w_correct_label=-1;
	
	for( j=1; j < nsum; j++ )
	{
		realnum current_error = (realnum)fabs(lines[j].wavelength()-wavelength);
		/* use pre-capitalized version of label to be like input chLineLabel */
		const char *chCaps = lines[j].chCLab();

		if( current_error < smallest_error )
		{
			index_of_closest = j;
			smallest_error = current_error;
		}
			
		if( current_error < smallest_error_w_correct_label &&
			 (strcmp(chCaps,chCARD) == 0) )
		{
			index_of_closest_w_correct_label = j;
			smallest_error_w_correct_label = current_error;
		}

		/* check wavelength and chLabel for a match */
		/*if( fabs(lines[j].wavelength- wavelength)/MAX2(DELTA,wavelength)<errorwave && 
			strcmp(chCaps,chCARD) == 0 ) */
		if( lgDEBUG && (current_error <= errorwave || 
			  fp_equal( wavelength + errorwave, lines[j].wavelength() ) ||
			  fp_equal( wavelength - errorwave, lines[j].wavelength() ))  
			 && strcmp(chCaps,chCARD) == 0 )
		{
			/* match, so set emiss to emissivity in line */
			/* and announce success by returning line index within stack */
			printf("Matched %s %15.8g %ld %18.11g %s\n",
					 chCaps,wavelength,j,lines[j].wavelength(),
					 lines[j].biglabel().c_str());
		}
	}




	// Didn't find line, handle error
	if( index_of_closest_w_correct_label == -1 ||
		smallest_error_w_correct_label > errorwave )
	{
		/* >>chng 05 dec 21, report closest line if we did not find exact match, note that
		 * exact match returns above, where we will return negative number of lines in stack */
		fprintf( ioQQQ," PROBLEM findline did not find line " );
		prt_line_err( ioQQQ, chCARD, wavelength );
		if( index_of_closest >= 0 )
		{
			fprintf( ioQQQ,"  The closest line (any label) was   \"%s\"\n", 
						lines[index_of_closest].label().c_str() );
			if( index_of_closest_w_correct_label >= 0 )
			{
				fprintf( ioQQQ,"  The closest with correct label was \"%s\"\n", 
							lines[index_of_closest_w_correct_label].label().c_str() );
				fprintf( ioQQQ,"  Error was %15.8g vs. tolerance %15.8g\n", 
							smallest_error_w_correct_label, errorwave );
			}
			else
				fprintf( ioQQQ,"\n  No line found with label \"%s\".\n", chCARD );
			fprintf( ioQQQ,"\n" );
		}
		else
		{
			fprintf( ioQQQ,".\n PROBLEM No close line was found\n" );
			TotalInsanity();
		}
		return -3;
	}

	if (lgDEBUG)
		fprintf(ioQQQ,"Identified %ld\n",index_of_closest_w_correct_label);


	return index_of_closest_w_correct_label;
}

/*************************************************************************
 *
 * cdEms_ip obtain the local emissivity for a line with known index
 *
 ************************************************************************/
void cdEmis(
	const LinSv* line,
	/* the vol emissivity of this line in last computed zone */
	double *emiss ,
	// intrinsic or emergent
	bool lgEmergent )
{
	DEBUG_ENTRY( "cdEmis()" );

	if (line)
		*emiss = line->emslin(lgEmergent);
	else
		*emiss = 0.;
}
